/**
 * C library for acessing TCB services.
 * 
 * author: Paulo Sousa (pjsousa@di.fc.ul.pt)
 * date: 21/03/2003
 */

#include "stdafx.h"
#include "tcb_services_library.h"
#include "private_functions_and_structures.h"
#include "Winsock.h"

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}


// This is an example of an exported variable
TCB_SERVICES_LIBRARY_API int nTcb_services_library=0;

// This is an example of an exported function.
TCB_SERVICES_LIBRARY_API int fnTcb_services_library(void)
{
	return 42;
}


//used for communication TCB DLL<->TCB through memory mapping
service_request_response_t *srrt=NULL;
//used for communication TCB DLL wait_info()<-TCB DistFD service 
//through memory mapping
fail_info_to_app_t *fi=NULL;
//event used by TCB to wait for our requests
HANDLE tcb_wait_event;
//event used for waiting for TCB answers
HANDLE dll_wait_event;
//event used for waiting for TCB distributed 
//timing failure detection service info
HANDLE wait_info_event;
//handles for memory mapped
HANDLE fmap,fmap1;
//indicates if there is a session opened
bool opened=FALSE;
//allows to detect if a session was opened by a different process
HANDLE session_event;

/**
 * Opens the TCB session. This method should be called
 * first of all. 
 * In this version, only 1 session is allowed per machine.
 * 
 * @return 0 if OK, -1 if TCB is stopped, 
 *					-2 if a session is already open,
 *					-3 if other error
 */
TCB_SERVICES_LIBRARY_API int open_tcb_session(void) {

	//create event for TCB to wait for our requests
	session_event = CreateEvent(NULL, TRUE, FALSE, L"session_event");
	if (session_event==NULL) {
		//_wperror(L"Error creating event\n");
		//Sleep(5000);
		return -3;
	}
	//if event is already created, a session is already being used
	if (GetLastError()==ERROR_ALREADY_EXISTS) {
		return -2;
	}

	//create event for TCB to wait for our requests
	tcb_wait_event = CreateEvent(NULL, TRUE, FALSE, L"tcb_wait_event");
	if (tcb_wait_event==NULL) {
		//_wperror(L"Error creating event\n");
		//Sleep(5000);
		return -3;
	}
	//create event for waiting for TCB answers
	dll_wait_event = CreateEvent(NULL, TRUE, FALSE, L"dll_wait_event");
	if (dll_wait_event==NULL) {
		//_wperror(L"Error creating event\n");
		//Sleep(5000);
		return -3;
	}
	//create event for waiting for TCB distributed 
	//timing failure detection service info
	wait_info_event = CreateEvent(NULL, TRUE, FALSE, L"wait_info_event");
	if (wait_info_event==NULL) {
		//_wperror(L"Error creating event\n");
		//Sleep(5000);
		return -3;
	}
	//if event isn't already created, TCB is not running
	if (GetLastError()!=ERROR_ALREADY_EXISTS) {
		return -1;
	}
	//unset event because it can be set (from a previous session)
	ResetEvent(wait_info_event);

	//create memory mapping TCB Services<->DLL
	fmap = CreateFileMapping(INVALID_HANDLE_VALUE, 
		NULL, PAGE_READWRITE, 0, sizeof(service_request_response_t), 
		L"tcb_memory_map");

	if (fmap==NULL) {
		//_wperror(L"Error creating memory mapping\n");
		//Sleep(5000);
		return -1;
	}
	
	//get pointer to memory map
	srrt = (service_request_response_t *) 
			MapViewOfFile(fmap, FILE_MAP_WRITE, 0, 0, 0);
	
	if (srrt==NULL) {
		//_wperror(L"Error mapping memory\n");
		//Sleep(5000);
		return -1;
	}
	//fill with zeros
	memset(srrt, 0, sizeof(service_request_response_t));


	//create a memory mapping to be used in DistFDService->TCB DLL waitInfo() comm
	HANDLE fmap1 = CreateFileMapping(INVALID_HANDLE_VALUE, 
		NULL, PAGE_READWRITE, 0, sizeof(fail_info_to_app_t), 
		L"wait_info_memory_map");
	if (fmap1==NULL) {
		//_wperror(L"Error creating memory mapping 2\n");
		return -1;
	}
	//associate the fail_info_to_app_t with the memory mapping
	fi =(fail_info_to_app_t *) 
			MapViewOfFile(fmap1, FILE_MAP_WRITE, 0, 0, 0);
	if (fi==NULL) {
		//_wperror("Error mapping memory 2\n");
		return -1;
	}
	//fill with zeros
	memset(fi, 0, sizeof(fail_info_to_app_t));

	//init socket_fd with an invalid value
	tcb_s_info.socket_fd = -1;

	opened=TRUE;

	return 0;	
}


/**
 * Closes the TCB session.
 * 
 * @return 0 if OK, -1 if error
 */
TCB_SERVICES_LIBRARY_API int close_tcb_session(void){
	//unmap memory
	if(!UnmapViewOfFile(srrt)){
		//_wperror(L"Error unmaping view of file\n");
		Sleep(5000);
		return -1;
	}
	//close memory map handle
	CloseHandle(fmap);
	//close event handles
	CloseHandle(dll_wait_event);
	CloseHandle(tcb_wait_event);
	opened=FALSE;
	return 0;
}


/**
 * Calls the tcb service: get_timestamp
 * 
 * Returns the local clock value (milliseconds).
 * 
 * @return Local clock value (milliseconds) if OK,
 *		   -1 if no TCB session is opened		   
 */
TCB_SERVICES_LIBRARY_API int get_tcb_timestamp(void){
	if (!opened) {
		printf("Open a TCB session first!\n");
		return -1;
	}

	//call GET_TIMESTAMP
	srrt->request.service_id=GET_TIMESTAMP;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int timestamp = srrt->response.value;
	//fill memory mapped struct with zeros
	//memset(srrt, 0, sizeof(service_request_response_t));
	srrt->type=0;
	//return answer
	return timestamp;
}


/**
 * Calls the tcb service: get_global_timestamp
 * 
 * Returns the global clock value (milliseconds) at local instant specified.
 * If localInstant is 0, the present global clock value is returned.
 * 
 * @param localInstant A timestamp obtained with get_timestamp() or 0.
 * 
 * @return Global clock value (milliseconds) at local instant specified if OK
 * 		   (If localInstant is 0, the present global clock value is returned),
 *			-1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API int get_tcb_global_timestamp(int localInstant){
	if (!opened) {
		printf("Open a TCB session first!\n");
		return -1;
	}
	//call GET_GLOBAL_TIMESTAMP
	srrt->request.service_id=GET_GLOBAL_TIMESTAMP;
	srrt->request.localInstant = localInstant;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int timestamp = srrt->response.value;
	//fill memory mapped struct with zeros
	//memset(srrt, 0, sizeof(service_request_response_t));
	srrt->type=0;
	//return answer
	return timestamp;
}


/**
 * Calls the tcb pseudo-service: get_tcb_id
 * 
 * Returns the id of the TCB we are connected to.
 * 
 * @return Id of the TCB we are connected to if OK,
 *		   -1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API char get_tcb_id(void){
	if (!opened) {
		printf("Open a TCB session first!\n");
		return -1;
	}
	//call GET_TCB_ID
	srrt->request.service_id=GET_TCB_ID;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	char id = srrt->response.id;
	//fill memory mapped struct with zeros
	//memset(srrt, 0, sizeof(service_request_response_t));
	srrt->type=0;
	//return answer
	return id;
}


/**
 * Calls the tcb service: start_local_measurement
 * 
 * Starts a local measurement at instant start_ev.
 * 
 * @param start_ev Instant that marks the begin of the measurement
 * 
 * @return Tag that identifies the measurement if OK,
 *		   -1 if no tag is available,
 *		   -2 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API int start_tcb_local_measurement(int start_ev){
	if (!opened) {
		printf("Open a TCB session first!\n");
		return -2;
	}
	//call START_MEASUREMENT
	srrt->request.service_id=START_MEASUREMENT;
	srrt->request.start_ev = start_ev;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int tag = srrt->response.tag;
	//fill memory mapped struct with zeros
	//memset(srrt, 0, sizeof(service_request_response_t));
	srrt->type=0;
	//return answer
	return tag;	
}


/**
 * Calls the tcb service: end_local_measurement
 * 
 * Stops the measurement identified by tag.
 * 
 * @param tag Measurement identifier
 *
 * @return Information about measurement if OK,
 *		   tcb_measurement_info_t.end_ev=-1 if tag is invalid,
 *		   tcb_measurement_info_t.end_ev=-2 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API tcb_measurement_info_t end_tcb_local_measurement(int tag){
	if (!opened) {
		printf("Open a TCB session first!\n");
		tcb_measurement_info_t ret;
		ret.end_ev = -2;
		return ret;
	}
	//call END_MEASUREMENT
	srrt->request.service_id=END_MEASUREMENT;
	srrt->request.tag = tag;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	
	//construct return value
	tcb_measurement_info_t ret;
	ret.end_ev = srrt->response.mi.end_ev;
	ret.procdelay = srrt->response.mi.procdelay;

	return ret;
}


/**
 * Calls the tcb service: timely_exec
 * 
 * Starts the execution of an application function. 
 * TCB verifies if it's possible to timely execute it.
 * 
 * @param start_ev A time instant
 * @param delay start_ev+delay marks the exec begin
 * @param max_exec Deadline for the execution, counting since start_ev
 * @param dllName Name of the DLL file where the function is defined
 * @param funcName Function name 
 * @param wcetName Worst Case Execution Time function name
 * 
 * @return The time instant when the function ends if OK;
 *			-1 if no available entry in the executions table
 *			-2 if dll not found
 *			-3 if function not found
 *			-4 if wcet function not found
 *			-5 if it's not possible to execute function
 *			-6 if it's not possible to timely execute function
 *			-7 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API int tcb_timely_exec(int start_ev, int delay, int max_exec, char* dllName,
										 char* funcName, char* wcetName){
	if (!opened) {
		printf("Open a TCB session first!\n");
		return -7;
	}
	//call TIMELY_EXEC
	srrt->request.service_id=TIMELY_EXEC;
	srrt->request.execArguments.start_ev=start_ev;
	srrt->request.execArguments.delay = delay;
	srrt->request.execArguments.max_exec = max_exec;
	//convert dllName to wchar_t
	wsprintf(srrt->request.execArguments.dllName, L"%S", dllName);
	//convert funcName to wchar_t
	wsprintf(srrt->request.execArguments.funcName, L"%S", funcName);
	//convert wcetName to wchar_t
	wsprintf(srrt->request.execArguments.wcetName, L"%S", wcetName);
	
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int end_ev = srrt->response.end_ev;
	srrt->type=0;
	//return answer
	return end_ev;
}


/**
 * Calls the tcb service: start_local_fd
 *
 * Starts a local timing failure detection.
 * 
 * @param start_ev A time instant
 * @param spec A timing failure occurs if start_ev+spec instant is reached
 *	            before a call endLocalFD
 * @param deadline Deadline for the execution of handler, if a failure occurs
 * @param dllName Name of the DLL file where the handler is defined
 * @param funcName Handler name 
 * @param wcetName Handler's Worst Case Execution Time function name
 *
 * @return Tag that identifies the timing failure detection if OK,
 *			-1 if no available entry in the localFD table,
 *			-2 if dll not found,
 *			-3 if function not found,
 *			-4 if wcet function not found,
 *			-5 if it's not possible to execute timing failure detection,
 *			-6 if a timing failure already occured,
 *			-7 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API int start_tcb_local_fd(int start_ev, int spec, int deadline, 
											char* dllName, char* funcName, char* wcetName){
	if (!opened) {
		printf("Open a TCB session first!\n");
		return -7;
	}
	//call START_LOCAL_FD
	srrt->request.service_id=START_LOCAL_FD;
	srrt->request.localFDArguments.start_ev=start_ev;
	srrt->request.localFDArguments.spec = spec;
	srrt->request.localFDArguments.deadline = deadline;
	//convert dllName to wchar_t
	wsprintf(srrt->request.localFDArguments.dllName, L"%S", dllName);
	//convert funcName to wchar_t
	wsprintf(srrt->request.localFDArguments.funcName, L"%S", funcName);
	//convert wcetName to wchar_t
	wsprintf(srrt->request.localFDArguments.wcetName, L"%S", wcetName);
    
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int tag = srrt->response.tag;
	srrt->type=0;
	//return answer
	return tag;
}


/**
 * Calls the tcb service: end_local_fd
 *
 * Ends a local timing failure detection.
 * 
 * @param tag Local timing detection identifier
 *
 * @return Information about the timing failure detection if OK,
 *			local_fd_info_t.end_ev=-1 if tag is invalid,
 *			local_fd_info_t.end_ev=-2 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API tcb_local_fd_info_t end_tcb_local_fd(int tag){
	if (!opened) {
		printf("Open a TCB session first!\n");
		tcb_local_fd_info_t ret;
		ret.end_ev=-2;
		return ret;
	}
	//call END_LOCAL_FD
	srrt->request.service_id=END_LOCAL_FD;
	srrt->request.tag = tag;
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);

	//construct return value
	tcb_local_fd_info_t ret;
	ret.end_ev = srrt->response.endLocalFDRet.end_ev;
	ret.faulty = srrt->response.endLocalFDRet.faulty;
	ret.procdelay = srrt->response.endLocalFDRet.procdelay;

	return ret;
}


/**
 * Waits for info from TCB distributed timing failure detection service.
 *
 * @return Information about the distributed timing failure detection if OK,
 *		   tcb_distributed_fd_info_t.from=-1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API tcb_distributed_fd_info_t tcb_wait_info(void){
	if (!opened) {
		printf("Open a TCB session first!\n");
		tcb_distributed_fd_info_t ret;
		ret.from = -1;
		return ret;
	}
	//wait for the info
	WaitForSingleObject(wait_info_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(wait_info_event);

	//construct DistributedFDInfo to return
	tcb_distributed_fd_info_t dfi;
	dfi.from = fi->from;
	dfi.seq = fi->seq;
	for (int i=0; i<NUMBER_OF_MACHINES; i++) {
		dfi.failInfo[i].fail = fi->failure_info[i].fail;
		dfi.failInfo[i].latenessAvail = fi->failure_info[i].lateness_avail;
		dfi.failInfo[i].latenessDegree = fi->failure_info[i].lateness_degree;
	}

	return dfi;
}


/**
 * Creates a TCB socket. 
 * At each time, there can be only one tcb socket created, so no descriptor 
 * is returned.
 *
 * @param family Address family specification
 *
 * @return 0 if OK, 
 *		   -1 if a socket error occurs.
 *		   -2 if no TCB session is opened.
 *		   
 */
TCB_SERVICES_LIBRARY_API int tcb_socket(int family) {
	if (!opened) {
		printf("Open a TCB session first!\n");
		return -2;
	}
	//create real socket
	tcb_s_info.socket_fd = socket(family, SOCK_DGRAM, 0);
	if (tcb_s_info.socket_fd>=0) {
		//get tcb id
		tcb_s_info.tcb_id = get_tcb_id();
		return 0;
	}
	return -1;
}


/**
 * Associates a local address with the TCB socket.
 *
 * @param addr Address to assign to the socket
 * @param addrlen Length of address
 *
 * @return 0 if OK, 
 *		   1 if no TCB session is opened,
 *		   SOCKET_ERROR if a socket error occurs
 */
TCB_SERVICES_LIBRARY_API int tcb_bind(struct sockaddr *addr, int addrlen) {
	if (!opened) {
		printf("Open a TCB session first!\n");
		return 1;
	}
	return bind(tcb_s_info.socket_fd, addr, addrlen);
}


/**
 * Calls the TCB pseudo-service: define_remote_fd_handler
 *
 * GOAL: to define the handler to be used on remote failure detection.
 *
 * PARAM:
 *		dllName Name of the DLL file where the handler is defined
 *		funcName Handler name 
 *		wcetName Handler's Worst Case Execution Time function name
 *		deadline Maximum exec time of failure handler (if a failure occurs)
 *
 * RETURNS: 
 *		0 if OK,
 *		-2 if dll not found,
 *		-3 if function not found,
 *		-4 if wcet function not found,
 *		-5 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API int define_tcb_remote_fd_handler(char* dllName, char* funcName, 
														  char* wcetName, int deadline){
	if (!opened) {
		printf("Open a TCB session first!\n");
		return -5;
	}
	//call DEFINE_REMOTE_HANDLER
	srrt->request.service_id=DEFINE_REMOTE_HANDLER;
	//convert dllName to wchar_t
	wsprintf(srrt->request.remote_handler.dllName, L"%S", dllName);    
	//convert funcName to wchar_t	
	wsprintf(srrt->request.remote_handler.funcName, L"%S", funcName);    
	//convert wcetName to wchar_t	
	wsprintf(srrt->request.remote_handler.wcetName, L"%S", wcetName);
    
	//set deadline
	srrt->request.remote_handler.deadline = deadline;
	
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int ret = srrt->response.ret;
	srrt->type=0;
	//return answer
	return ret;		
}


/**
 * Sends data to a specific destination through the TCB socket,
 * starting a distributed measurement.
 *
 * @param buf buffer that contains the data to be transmitted. 
 *			  ATTENTION: The last TCB_DATA_SIZE bytes of the buffer 
 *						 will be used by TCB
 * @param len length of buf
 * @param flags specifies the way in which the call is made
 * @param to address of the target socket
 * @param tolen length of address
 * @param send_ev the initial time instant of the action to be measured
 *
 * @return total number of bytes sent if OK, 
 *		   -100 if no TCB session is opened,
 *		   SOCKET_ERROR if a socket error occurs
 */
TCB_SERVICES_LIBRARY_API int tcb_sendto(char* buf, int len, int flags, 
										struct sockaddr *to, int tolen, int send_ev) {
	if (!opened){
		printf("Open a TCB session first!\n");
		return -100;
	}
	
	return my_send(buf, len, flags, to, tolen, send_ev, 0);
}


/**
 * Sends data to a specific destination through the TCB socket,
 * starting simultaneously a distributed measurement and a distributed
 * timing failure detection.
 * If a timing failure occurs, the failure handler is executed in the sender.
 *
 * If data is not delivered to all recipients specified in destMask until send_ev+spec,
 * the specified failure handler is executed (in the sender) with the maximum
 * exec time specified.
 *
 * @param buf buffer that contains the data to be transmitted. 
 *			  ATTENTION: The last TCB_DATA_SIZE bytes of the buffer 
 *						 will be used by TCB
 * @param len length of buf
 * @param flags specifies the way in which the call is made
 * @param to address of the target socket
 * @param tolen length of address
 * @param send_ev the initial time instant of the failure detection and measurement
 * @param dest_mask bit mask indicating packet destinations. If bit n=1,
 *				    then packet is sent to destination where TCB id=n.
 *				    E.g. dest_mask=0x7 => send to destinations with TCB ids=0,1,2.
 *					This information should be consistent with the remote IP address
 *					specified in to
 * @param spec Maximum duration starting at send_ev (for failure detection)
 * @param deadline Maximum exec time of failure handler (if a failure occurs)
 * @param dllName Name of the DLL file where the failure handler is defined
 * @param funcName Failure handler name
 * @param wcetName Failure handler Worst Case Execution Time function name
 *
 * @return total number of bytes sent if OK, 
 *		   -100 if no TCB session is opened,
 *			SOCKET_ERROR if a socket error occurs
 *			-2 if dll not found,
 *			-3 if function not found,
 *			-4 if wcet function not found,
 *			-5 if there are too many distributed fds already running,
 *			-6 if a timing failure already occured,
 *			-7 if it's not possible to execute timing failure detection
 */
TCB_SERVICES_LIBRARY_API int tcb_sendtoWithFD(char* buf, int len, int flags, 
										struct sockaddr *to, int tolen, int send_ev,
										char dest_mask, int spec, int deadline,
										char* dllName, char* funcName, char* wcetName) {
	if (!opened){
		printf("Open a TCB session first!\n");
		return -100;
	}

	int seq=0;
	//if there is at least one destination, call start_distributed_fd
	if (dest_mask!=0){
		seq = start_tcb_distributed_fd(dest_mask, send_ev, spec, deadline, 
										dllName, funcName, wcetName);
		if (seq<0)
			return seq;

		//serialize TCB id
		buf[len-TCB_DATA_SIZE+TCB_ID_OFFSET]=tcb_s_info.tcb_id;
	}

	return my_send(buf, len, flags, to, tolen, send_ev, seq);
}


/**
 * Sends data to a specific destination through the TCB socket,
 * starting simultaneously a distributed measurement and a distributed
 * timing failure detection.
 * If a timing failure occurs, the failure handler is executed in the recipients.
 *
 * For each recipient specified in dest_mask: if data is not delivered until send_ev+spec,
 * the failure handler specified in define_tcb_remote_fd_handler
 * is executed with the maximum exec time specified in that function.
 *
 * @param buf buffer that contains the data to be transmitted. 
 *			  ATTENTION: The last TCB_DATA_SIZE bytes of the buffer 
 *						 will be used by TCB
 * @param len length of buf
 * @param flags specifies the way in which the call is made
 * @param to address of the target socket
 * @param tolen length of address
 * @param send_ev the initial time instant of the failure detection and measurement
 * @param dest_mask bit mask indicating packet destinations. If bit n=1,
 *				    then packet is sent to destination where TCB id=n.
 *				    E.g. dest_mask=0x7 => send to destinations with TCB ids=0,1,2.
 *					This information should be consistent with the remote IP address
 *					specified in to
 * @param spec Maximum duration starting at send_ev (for failure detection)
 *
 * @return total number of bytes sent if OK, 
 *		   -100 if no TCB session is opened,
 *			SOCKET_ERROR if a socket error occurs,
 *			-5 if there are too many distributed fds already running,
 *			-6 if a timing failure already occured,
 *			-7 if it's not possible to execute timing failure detection
 */
TCB_SERVICES_LIBRARY_API int tcb_sendtoWithRemoteFD(char* buf, int len, int flags, 
										struct sockaddr *to, int tolen, int send_ev,
										char dest_mask, int spec) {
	if (!opened){
		printf("Open a TCB session first!\n");
		return -100;
	}

	int seq=0;
	//if there is at least one destination, call start_distributed_remote_fd
	if (dest_mask!=0){
		seq = start_tcb_distributed_remote_fd(dest_mask, send_ev, spec);
		//if an error occured, return now
		if (seq<0)
			return seq;

		//serialize TCB id
		buf[len-TCB_DATA_SIZE+TCB_ID_OFFSET]=tcb_s_info.tcb_id;
	}
	//send msg
	return my_send(buf, len, flags, to, tolen, send_ev, seq);
}


/**
 * Receives data from the TCB socket. This method ends a distributed
 * measurement but doesn't end a distributed failure detection. For that, 
 * call end_tcb_distributed_action() with the tcb_distributed_fd_id_t returned by 
 * this method.
 * 
 * @param buf buffer for the incoming data.
 *			  ATTENTION: The last TCB_DATA_SIZE bytes of the buffer 
 *						 will be used by TCB
 * @param len (IN:length of buf)(OUT:bytes received)
 * @param flags specifies the way in which the call is made
 * @param from pointer to a buffer that will hold the source address upon return
 * @param fromlen (IN:pointer to the size of the from buffer)
 *				  (OUT:actual size of from buffer)
 * 
 * @return The return value is a union. Its value depends on the service 
 *		   called by the sender. 
 *		   If it's only a distributed measurement, a 
 *		   tcb_distributed_measurement_info_t is returned.
 * 		   If it's a distributed failure detection, a 
 *		   tcb_distributed_fd_id_t is returned.
 *
 *		   If there is no TCB session opened, 
 *		   tcb_distributed_action_result_t.type=-100.
 *         
 *		   If a socket error, tcb_distributed_action_result_t.type=SOCKET_ERROR.
 * 
 */
TCB_SERVICES_LIBRARY_API tcb_distributed_action_result_t tcb_recvfrom(
										  char* buf, int* len, int flags, 
										  struct sockaddr* from, int* fromlen) {
	
	tcb_distributed_action_result_t ret;
	if (!opened){
		printf("Open a TCB session first!\n");
		ret.type = -100;
		return ret;
	}

	//receive msg
	int nbytes=recvfrom(tcb_s_info.socket_fd, buf, *len, flags, from, fromlen);
	//if error, return now
	if (nbytes<=0) {
		if (nbytes==0)
			nbytes=-1;
		ret.type = nbytes;
		return ret;
	}
	*len = nbytes;

	//extract send_ev
	int send_ev = byteArrayToInt(buf, nbytes-TCB_DATA_SIZE+SEND_EV_OFFSET);

	bool deliver=TRUE;
	//extract sequence number
	int seq = byteArrayToInt(buf, nbytes-TCB_DATA_SIZE+SEQ_OFFSET);
	//check if msg was sent with or without FD
	if (seq!=0) {
		//msg was sent with FD
		//pre-allocate space to guarantee that end_tcb_distributed_action() succeeds.
		char tcb_id = buf[nbytes-TCB_DATA_SIZE+TCB_ID_OFFSET];
		deliver = tcb_msg_rcv_on_payload(seq, tcb_id, -2);
		if (deliver) {
			ret.type=1;
			ret.fd_info.send_ev = send_ev;
			ret.fd_info.seq = seq;
			ret.fd_info.src_id = tcb_id;
			return ret;
		}
		//If there is no free space, don't deliver message
		ret.type= -1;
		return ret;
	}
	//if sent without FD, end measurement
	ret.type=0;
	//get a timestamp from TCB
	ret.measurement_info.delivery_ev = get_tcb_timestamp();
	//get the global timestamp to calculate the msg delay
	int delivery_ev_global = get_tcb_global_timestamp(ret.measurement_info.delivery_ev);
	//calculate the delay
	ret.measurement_info.delay = delivery_ev_global- send_ev;
	//error is related with clock synchronization algorithm
	ret.measurement_info.error=DISTRIBUTED_MEASUREMENT_ERROR;
	return ret;			
}


/**
 * Ends a distributed failure detection and the associated measurement.
 * 
 * @param dfi Id of the action to be ended.
 * 
 * @return Info about the result of the distributed measurement if OK,
 *		   tcb_distributed_measurement_info_t.end_ev=-100 if there is no 
 *		   TCB session opened,
 *		   -1 if a socket error occurs.
 */
TCB_SERVICES_LIBRARY_API tcb_distributed_measurement_info_t 
			end_tcb_distributed_action(tcb_distributed_fd_id_t dfi) {
				
	tcb_distributed_measurement_info_t ret;
	if (!opened){
		printf("Open a TCB session first!\n");
		ret.delivery_ev = -100;
		return ret;
	}

	//get a timestamp from TCB
	ret.delivery_ev = get_tcb_timestamp();
	//get the global timestamp to calculate the msg delay
	int delivery_ev_global = get_tcb_global_timestamp(ret.delivery_ev);
	//calculate the delay
	ret.delay = delivery_ev_global - dfi.send_ev;
	ret.error=DISTRIBUTED_MEASUREMENT_ERROR;
	if (dfi.seq!=0) {
		boolean ok;
		//if msg was sent with FD, notify TCB
		ok=tcb_msg_rcv_on_payload(dfi.seq, dfi.src_id, ret.delivery_ev);
		if (!ok) {
			ret.delivery_ev = -1;
			return ret;
		}
	}
	return ret;
}
	



/*******************
 * PRIVATE FUNCTIONS
 *******************/

/**
 * Private function that sends data through TCB socket.
 * Called by the various public sendTo().
 *
 * @param buf buffer that contains the data to be transmitted. 
 *			  ATTENTION: The last TCB_DATA_SIZE bytes of the buffer 
 *						 will be used by TCB
 * @param len length of buf
 * @param flags specifies the way in which the call is made
 * @param to address of the target socket
 * @param tolen length of address
 * @param send_ev the initial time instant of the action to be measured
 * @param seq sequence number to serialize in msg
 *
 * @return total number of bytes sent if OK, 
 *		   SOCKET_ERROR if a socket error occurs
 */
int my_send (char* buf, int len, int flags, struct sockaddr *to, int tolen, 
			 int send_ev, int seq) {
	//get global timestamp corresponding to local send_ev
	int send_ev_global = get_tcb_global_timestamp(send_ev);
	
	//serialize send_ev
	intToByteArray(&send_ev_global, buf, len-TCB_DATA_SIZE+SEND_EV_OFFSET);
	//serialize seq
	intToByteArray(&seq, buf, len-TCB_DATA_SIZE+SEQ_OFFSET);
	//send msg
	return sendto(tcb_s_info.socket_fd, buf, len, flags, to, tolen);
}





/**
 * Calls the tcb service: start_distributed_fd
 * 
 * Starts a distributed timing failure detection.
 * 
 * @param destMask Bit mask indicating packet destinations.
 * 					if (bit n == 1) then pack is sent to machine n.
 * 					E.g. destMask=0x7 => send to machines 0,1,2
 * @param sendEv The initial time instant of the failure detection.
 * @param spec Maximum duration starting at sendEv.
 * @param deadline Maximum exec time of failure handler (if a failure occurs)
 * @param dllName Name of the DLL file where the handler is defined
 * @param funcName Handler name 
 * @param wcetName Handler's Worst Case Execution Time function name
 *
 * @return seq that identifies the dist. timing failure detection,
 *			-2 if dll not found,
 *			-3 if function not found,
 *			-4 if wcet function not found,
 *			-5 if there are too many distributed fds already running,
 *			-6 if a timing failure already occured,
 *			-7 if it's not possible to execute timing failure detection. 
 */ 
int start_tcb_distributed_fd(char destMask, int sendEv, int spec, 
						 int deadline, char* dllName, char* funcName, 
						 char* wcetName) {
	//call START_DISTRIBUTED_FD
	srrt->request.service_id=START_DISTRIBUTED_FD;
	srrt->request.distFDArguments.destMask = destMask;
	srrt->request.distFDArguments.send_ev = sendEv;
	srrt->request.distFDArguments.spec = spec;
	srrt->request.distFDArguments.deadline = deadline;
	//convert dllName to wchar_t
	wsprintf(srrt->request.distFDArguments.dllName, L"%S", dllName);
	//convert funcName to wchar_t
	wsprintf(srrt->request.distFDArguments.funcName, L"%S", funcName);    
	//convert wcetName to wchar_t
	wsprintf(srrt->request.distFDArguments.wcetName, L"%S", wcetName);
    	
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int seq = srrt->response.seq;
	srrt->type=0;
	//return answer
	return seq;		
}


/**
 * Calls the tcb service: start_distributed_remote_fd
 * 
 * Starts a distributed remote timing failure detection.
 * 
 * @param destMask Bit mask indicating packet destinations.
 * 					if (bit n == 1) then pack is sent to machine n.
 * 					E.g. destMask=0x7 => send to machines 0,1,2
 * @param sendEv The initial time instant of the failure detection.
 * @param spec Maximum duration starting at sendEv.
 *
 * @return seq that identifies the dist. timing failure detection,
 *			-5 if there are too many distributed fds already running,
 *			-6 if a timing failure already occured,
 *			-7 if it's not possible to execute timing failure detection. 
 */ 
int start_tcb_distributed_remote_fd(char destMask, int sendEv, int spec) {
	//call START_DISTRIBUTED_REMOTE_FD
	srrt->request.service_id=START_DISTRIBUTED_REMOTE_FD;
	srrt->request.distFDArguments.destMask = destMask;
	srrt->request.distFDArguments.send_ev = sendEv;
	srrt->request.distFDArguments.spec = spec;
		
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	int seq = srrt->response.seq;
	srrt->type=0;
	//return answer
	return seq;		
}


/**
 * PSEUDO-SERVICE: msg_rcv_on_payload
 *
 * GOAL: to DLL notify TCB that a msg was received on payload
 * 
 * PARAM: 
 *		seq - sequence number
 *		from - sender id
 *		time_of_receipt - reception time instant
 *
 * RETURNS : TRUE if msg can be delivered,
 *					  FALSE if msg cannot be delivered			   
 */
bool tcb_msg_rcv_on_payload(int seq, char from, int time_of_receipt){
	//call MSG_RCV_ON_PAYLOAD
	srrt->request.service_id=MSG_RCV_ON_PAYLOAD;
	srrt->request.msgRcvOnPayloadArguments.seq=seq;
	srrt->request.msgRcvOnPayloadArguments.from=from;
	srrt->request.msgRcvOnPayloadArguments.time_of_receipt=time_of_receipt;
		
	//it's a request
	srrt->type=1;
	//sinalize TCB services thread that a request is ready
	SetEvent(tcb_wait_event);
	//wait for an answer
	WaitForSingleObject(dll_wait_event, INFINITE);
	//after receiving the answer, set event to false
	ResetEvent(dll_wait_event);
	bool deliver = srrt->response.deliver;
	srrt->type=0;
	//return answer
	return deliver;		
}


/*******************************
 * Auxiliary Functions
 *******************************/

/**
 * Desserializes a char array containing an integer.
 * 
 * @param b the char array to desserialize
 * @param off start offset
 * @return the desserialized integer.
 */
int byteArrayToInt(char* b, int off) {
	int ret = 0;
	char* ret_pt = (char*)&ret;

	ret_pt[0] = b[off];
	ret_pt[1] = b[off + 1]; 
	ret_pt[2] = b[off + 2]; 
	ret_pt[3] = b[off + 3];

	return ret;
}


/**
 * Serializes an integer into a char array
 * 
 * @param i the int to serialize
 * @param b to return the serialization
 * @param off start offset
 */
void intToByteArray(int* i, char* b, int off) {
	
	char* int_pt = (char*)i;
	

	b[off+0] = int_pt[0];
	b[off+1] = int_pt[1];
	b[off+2] = int_pt[2];
	b[off+3] = int_pt[3];
}

